import os

from sqlalchemy import create_engine, Column, Integer, String, ForeignKey, DateTime, Text, Boolean, Float, Index
from sqlalchemy.orm import declarative_base, relationship, Session

from cuwais.config import config_file

_Base = declarative_base()
_Engine = create_engine(config_file.get("db_connection"), echo=config_file.get("debug"), future=True)


def create_session() -> Session:
    return Session(_Engine)


class User(_Base):
    __tablename__ = 'user'

    id = Column(Integer, primary_key=True)
    real_name = Column(Text, nullable=False)
    nickname = Column(Text, nullable=False, unique=True, index=True)
    google_id = Column(String(255), nullable=True, unique=True, index=True)
    is_bot = Column(Boolean, unique=False, nullable=False, default=False)
    is_admin = Column(Boolean, unique=False, nullable=False, default=False)
    display_real_name = Column(Boolean, unique=False, nullable=False, default=False)

    submissions = relationship("Submission", back_populates="user")

    @property
    def display_name(self):
        return self.nickname if not self.display_real_name else self.real_name

    def to_public_dict(self) -> dict:
        public_vals = {'_cuwais_type': 'user',
                       'user_id': self.id,
                       'display_name': self.display_name,
                       'nickname': self.nickname,
                       'display_real_name': self.display_real_name,
                       'is_bot': self.is_bot,
                       'is_admin': self.is_admin}

        if self.display_real_name:
            public_vals['real_name'] = self.real_name

        return public_vals

    def to_private_dict(self) -> dict:
        private_vals = {'real_name': self.real_name,
                        'google_id': self.google_id}

        return {**self.to_public_dict(), **private_vals}

    def __repr__(self):
        return f"User(id={self.id!r}, nickname={self.nickname!r})"


class Submission(_Base):
    __tablename__ = 'submission'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=False, index=True)
    submission_date = Column(DateTime, unique=False, nullable=False)
    url = Column(Text, unique=False, nullable=False)
    active = Column(Boolean, unique=False, nullable=False, default=True, index=True)
    files_hash = Column(Text, unique=False, nullable=False)

    user = relationship("User", back_populates="submissions")
    results = relationship("Result", back_populates="submission")

    def to_public_dict(self) -> dict:
        return {'_cuwais_type': 'submission',
                'submission_id': self.id,
                'user_id': self.user_id,
                'submission_date': self.submission_date}

    def to_private_dict(self) -> dict:
        private_vals = {'url': self.url,
                        'active': self.active,
                        'files_hash': self.files_hash}

        return {**self.to_public_dict(), **private_vals}

    def __repr__(self):
        return f"Submission(id={self.id!r}, url={self.url!r}, url={self.submission_date!r})"


class Match(_Base):
    __tablename__ = 'match'

    id = Column(Integer, primary_key=True)
    match_date = Column(DateTime, unique=False, nullable=False, index=True)
    recording = Column(Text, unique=False, nullable=False)

    results = relationship("Result", back_populates="match")

    def to_public_dict(self) -> dict:
        return {'_cuwais_type': 'match',
                'match_id': self.id,
                'match_date': self.match_date,
                'recording': self.recording}

    def to_private_dict(self) -> dict:
        private_vals = {}

        return {**self.to_public_dict(), **private_vals}

    def __repr__(self):
        return f"Match(id={self.id!r}, match date={self.match_date!r})"


class Result(_Base):
    __tablename__ = 'result'

    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('match.id'), nullable=False, index=True)
    submission_id = Column(Integer, ForeignKey('submission.id'), nullable=False, index=True)
    outcome = Column(Integer, unique=False, nullable=False, index=True)
    points_delta = Column(Float, unique=False, nullable=False)
    healthy = Column(Boolean, unique=False, nullable=False, index=True)
    player_id = Column(Text, unique=False, nullable=False)
    prints = Column(Text, unique=False, nullable=True)
    result_code = Column(Text, unique=False, nullable=False)

    submission = relationship("Submission", back_populates="results")
    match = relationship("Match", back_populates="results")

    def to_public_dict(self) -> dict:
        return {'_cuwais_type': 'result',
                'match_id': self.match_id,
                'submission_id': self.submission_id,
                'outcome': self.outcome,
                'player_id': self.player_id}

    def to_private_dict(self) -> dict:
        private_vals = {'points_delta': self.points_delta,
                        'healthy': self.healthy,
                        'prints': self.prints,
                        'result_code': self.result_code}

        return {**self.to_public_dict(), **private_vals}

    def __repr__(self):
        return f"Result(id={self.id!r}, match id={self.match_id!r}, submission id={self.submission_id!r})"


Index('healthy_outcome_index', Result.healthy, Result.outcome)


def create_tables():
    _Base.metadata.create_all(_Engine)



import logging
from typing import Optional, Union

import yaml
import os


def _fill_in_defaults(data: Optional[Union[dict, list]], defaults: Optional[Union[dict, list]]):
    if data is None:
        return defaults

    if isinstance(data, list):
        return data

    if isinstance(data, dict):
        if not isinstance(defaults, dict):
            return data

        for key in data:
            data[key] = _fill_in_defaults(data[key], defaults.get(key, None))

        for key in defaults:
            if key not in data:
                data[key] = defaults[key]

        return data

    return data


class Config:
    def __init__(self, path: str, defaults: Optional[Union[dict, list, 'Config']] = None):
        if isinstance(defaults, Config):
            defaults = defaults.get_all()

        self._defaults = defaults

        if not os.path.isfile(path):
            logging.error(f"Could not find config file {path}")
            self._data = None
        else:
            with open(path, 'r') as stream:
                self._data = yaml.safe_load(stream)

        if self._defaults is None:
            self._defaults = dict()
        if self._data is None:
            self._data = dict()

        self._data = _fill_in_defaults(self._data, defaults)

    def get_all(self):
        return self._data

    def get(self, key: str):
        if not isinstance(key, str):
            raise RuntimeError("Key should be of type str")

        parts = key.split(".")

        data = self._data
        for part in parts:
            if data is None:
                break
            if isinstance(data, dict):
                data = data.get(part, None)
            elif isinstance(data, list):
                i = int(part)
                data = None if i >= len(data) or i < 0 else data[i]

        return data


default_config_file = Config("default_config.yml")
config_file = Config("config.yml", default_config_file)



import hashlib
from enum import Enum, unique


@unique
class Outcome(Enum):
    Win = 1
    Loss = 2
    Draw = 3


@unique
class Result(Enum):
    ValidGame = "valid-game"
    Exception = "exception"
    IllegalMove = "illegal-move"
    IllegalBoard = "illegal-board"
    BrokenEntryPoint = "broken-entry-point"
    UnknownResultType = "unknown-result-type"
    GameUnfinished = "game-unfinished"
    Timeout = "timeout"
    ProcessKilled = "process-killed"


def calculate_git_hash(user_id: int, commit_hash: str, url: str) -> str:
    m = hashlib.sha256()
    m.update(str(user_id).encode('utf-8'))
    m.update(commit_hash.encode('utf-8'))
    # m.update(url.encode('utf-8'))
    return str(m.hexdigest())



import abc
import random
from typing import List

import chess as chess

from cuwais.config import config_file


class Gamemode:
    def __init__(self, name: str, players: List[str], options: dict):
        self._name = str(name)
        self._players = list(players)
        self._options = {"turn_time": 10, **dict(options)}

    @property
    def player_count(self):
        return len(self._players)

    @property
    def name(self):
        return self._name

    @property
    def players(self):
        return self._players.copy()

    @property
    def options(self):
        return self._options.copy()

    @abc.abstractmethod
    def setup(self, **options):
        """Gets the board that will be used by a game instance"""
        pass

    def filter_board(self, board, player: int):
        """Used to hide parts of the board, or transform the board into a form used by the player given"""
        return board

    def parse_move(self, move):
        """Optionally used to parse the returned move, eg if the AI can return a string or an object"""
        return move

    @abc.abstractmethod
    def is_move_legal(self, board, move) -> bool:
        """Checks if a move can be made with a given board"""
        pass

    @abc.abstractmethod
    def apply_move(self, board, move):
        """Applies the move to the board and returns the new board"""
        pass

    @abc.abstractmethod
    def is_win(self, board, player):
        """Checks if the board is in a won position given the player just moved"""
        pass

    @abc.abstractmethod
    def is_loss(self, board, player):
        """Checks if the board is in a lost position given the player just moved"""
        pass

    @abc.abstractmethod
    def is_draw(self, board, player):
        """Checks if the board is in a drawn position given the player just moved"""
        pass

    @abc.abstractmethod
    def encode_move(self, move, player) -> str:
        """Converts a move to a string form for storing or transmitting"""
        pass

    @abc.abstractmethod
    def encode_board(self, board) -> str:
        """Converts a board to a string form for storing or transmitting"""
        pass

    @staticmethod
    def get(gamemode_name):
        return {"chess": ChessGamemode()}.get(gamemode_name, None)

    @staticmethod
    def get_from_config():
        gamemode = Gamemode.get(config_file.get("gamemode.id").lower())
        options = config_file.get("gamemode.options")
        return gamemode, options


class ChessGamemode(Gamemode):
    def __init__(self):
        super(ChessGamemode, self).__init__(name="chess", players=["white", "black"], options={"chess_960": True})

    def setup(self, chess_960, **kwargs):
        if chess_960:
            return chess.Board.from_chess960_pos(random.randint(0, 959))

        return chess.Board()

    def is_win(self, board, move):
        return board.is_checkmate()

    def is_draw(self, board, move):
        return board.is_stalemate() or board.is_insufficient_material() or board.is_seventyfive_moves()

    def is_loss(self, board, move):
        return False

    def parse_move(self, move):
        if isinstance(move, str):
            move = chess.Move.from_uci(move)
        return move

    def is_move_legal(self, board, move):
        if move is None:
            return False
        if not isinstance(move, chess.Move):
            return False
        if move not in board.legal_moves:
            return False
        return True

    def apply_move(self, board: chess.Board, move: chess.Move):
        board.push(move)
        return board

    def encode_move(self, move: chess.Move, player: int) -> str:
        return move.uci()

    def encode_board(self, board: chess.Board) -> str:
        return board.fen()
